from qtpy import QtCore, QtWidgets

from functools import partial
import re

from .FileBrowser import FileBrowser


class TableRowEditor(QtWidgets.QDialog):
    data_collected = QtCore.Signal(list)

    def __init__(self, header: list[str], data: list[str] | None = None, name: str | None = None,
                 item_choices: list[list[str] | None] | None = None, regex_pattern: list[str | None] | None = None,
                 item_filter_list: list[list[str]] | None = None, is_browser: list[bool] | None = None, parent=None):
        self._header = header
        action = 'Add' if data is None else 'Edit'

        if data is None:
            data = [None] * len(header)
        if item_choices is None:
            item_choices = [None] * len(header)
        if is_browser is None:
            is_browser = [False] * len(header)

        self._old_data = data
        self._item_choices = item_choices
        self._regex_pattern = regex_pattern
        self._filter_list = item_filter_list
        self._is_browser = is_browser

        self._button_box: QtWidgets.QDialogButtonBox | None = None
        self._row_items: list[tuple[QtWidgets.QLabel, QtWidgets.QLineEdit | QtWidgets.QComboBox | FileBrowser]] | None = None

        super().__init__(parent=parent)
        if name is None:
            name = 'Row'
        self.setWindowTitle(f"{action} {name}")

        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Maximum)

        self.init_ui()
        self.set_layout()
        self.populate()

    @property
    def _new_data(self) -> list[str]:
        data = []
        for row_item in self._row_items:
            text = self.get_item_editor_text(row_item[1])
            data.append(text)
        return data

    @staticmethod
    def get_item_editor_text(item_editor) -> str:
        if hasattr(item_editor, 'path'):
            return item_editor.path
        if hasattr(item_editor, 'currentText'):
            return item_editor.currentText()
        else:
            return item_editor.text()

    def _create_row_items(self):
        editor_width = 720 if sum(self._is_browser) else 360
        self._row_items = []
        for i, (item_label, item_value, item_choices) in enumerate(zip(self._header, self._old_data, self._item_choices)):
            label = QtWidgets.QLabel(f"{item_label}:", self)
            label.setFixedWidth(120)
            if self._is_browser[i]:
                item_editor = FileBrowser(default_path=item_value if item_value else None, line_edit_width=0, parent=self)
            elif item_choices:
                item_editor = QtWidgets.QComboBox(self)
                item_editor.addItems(item_choices)
                if item_value:
                    item_editor.setCurrentIndex(item_choices.index(item_value))
            else:
                item_editor = QtWidgets.QLineEdit(item_value if item_value else "", self)
            if self._regex_pattern or self._filter_list:
                if hasattr(item_editor, 'textChanged'):
                    item_editor.textChanged.connect(self.validate_text)
            item_editor.setFixedWidth(editor_width)
            self._row_items.append((label, item_editor))

        # in case we match against an empty string
        for i, item_value in enumerate(self._old_data):
            item_editor = self._row_items[i][1]
            if hasattr(item_editor, 'textChanged'):
                item_editor.textChanged.emit(item_value if item_value else "")

    def init_ui(self):
        self._button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel, self)
        self._create_row_items()

        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        for row_item in self._row_items:
            sub_layout = QtWidgets.QHBoxLayout()
            sub_layout.addWidget(row_item[0])
            sub_layout.addWidget(row_item[1])
            self.layout().addLayout(sub_layout)

        self.layout().addWidget(self._button_box)

    @QtCore.Slot()
    def validate_text(self):
        ok_button_enabled = True
        for idx, row_item in enumerate(self._row_items):
            text = self.get_item_editor_text(row_item[1])

            use_regex = self._regex_pattern and self._regex_pattern[idx] is not None
            use_filter_list = self._filter_list and self._filter_list[idx] is not None
            if not (use_regex or use_filter_list):
                continue

            if (use_regex and re.match(self._regex_pattern[idx], text)) or (use_filter_list and text in self._filter_list[idx]):
                ok_button_enabled = False
                break

        ok_button = self._button_box.button(QtWidgets.QDialogButtonBox.Ok)
        ok_button.setEnabled(ok_button_enabled)

    def accept(self):
        self.data_collected.emit(self._new_data)
        super().accept()


class ParamTable(QtWidgets.QWidget):
    data_collected = QtCore.Signal(list, list)

    def __init__(self, header: list[str], data: list[list[str]], is_unique: list[bool] | None = None,
                 remember_deleted=True, parent=None, **kwargs):
        self._header = header
        self._old_data = data
        self._row_editor_kwargs = kwargs

        if is_unique is None:
            is_unique = [False] * len(header)
        self._is_unique = is_unique
        self._remember_deleted = remember_deleted

        self._add_button: QtWidgets.QPushButton | None = None
        self._delete_button: QtWidgets.QPushButton | None = None
        self._edit_button: QtWidgets.QPushButton | None = None
        self._reset_button: QtWidgets.QPushButton | None = None

        self._table: QtWidgets.QTableWidget | None = None
        self._table_items: list[list[QtWidgets.QTableWidgetItem]] | None = None
        self._is_deleted: list[bool] | None = None

        super().__init__(parent=parent)

        self.init_ui()
        self.set_layout()
        self.populate()

        self.setFocusPolicy(QtCore.Qt.StrongFocus)

    @property
    def _new_data(self):
        data = []
        for row in self._table_items:
            data_row = []
            for item in row:
                data_row.append(str(item.data(0)))
            data.append(data_row)
        return data

    @property
    def _current_row(self):
        return self._table.currentRow()

    def _get_item_filter_list(self, include_current_row=False):
        filter_list = [[] for _ in range(len(self._header))]
        for i, row in enumerate(self._table_items):
            if i == self._current_row and not include_current_row:
                continue
            for j, item in enumerate(row):
                if self._is_unique[j]:
                    filter_list[j].append(str(item.data(0)))
                else:
                    filter_list[j].append(None)
        return filter_list

    def _create_table_items(self):
        table_items = []
        for data_row in self._old_data:
            row = []
            for data_item in data_row:
                row.append(QtWidgets.QTableWidgetItem(data_item))
            table_items.append(row)

        self._table_items = table_items
        self._is_deleted = [False] * len(table_items)

    def _set_table_items(self):
        self._table.setRowCount(len(self._table_items))
        for i, row in enumerate(self._table_items):
            for j, item in enumerate(row):
                self._table.setItem(i, j, item)

    def init_ui(self):
        self._add_button = QtWidgets.QPushButton('Add...', self)
        self._delete_button = QtWidgets.QPushButton('Delete', self)
        self._edit_button = QtWidgets.QPushButton('Edit...', self)
        self._reset_button = QtWidgets.QPushButton('Reset', self)

        for w in (self._add_button, self._delete_button, self._edit_button, self._reset_button):
            w.setFixedWidth(90)
        self._set_buttons_enabled(False)

        self._table = QtWidgets.QTableWidget(len(self._old_data), len(self._header), self)
        self._table.setTextElideMode(QtCore.Qt.ElideLeft)
        self._table.horizontalHeader().setStretchLastSection(True)
        self._table.setHorizontalHeaderLabels(self._header)

        self._table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self._table.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self._table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self._table.setMinimumHeight(250)

        self._create_table_items()
        self._set_table_items()

        self._table.selectionModel().selectionChanged.connect(self.selection_changed_action)

        self._add_button.clicked.connect(self._add_row_dialog)
        self._delete_button.clicked.connect(self._delete_row)
        self._edit_button.clicked.connect(self._edit_row_dialog)
        self._reset_button.clicked.connect(self._reset)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        sub_layout = QtWidgets.QHBoxLayout()
        for w in (self._add_button, self._delete_button, self._edit_button, self._reset_button):
            sub_layout.addWidget(w)
        sub_layout.setAlignment(QtCore.Qt.AlignLeft)
        self.layout().addLayout(sub_layout)

        self.layout().addWidget(self._table)

    def _set_buttons_enabled(self, a0: bool):
        for w in (self._delete_button, self._edit_button):
            w.setEnabled(a0)

    @QtCore.Slot()
    def selection_changed_action(self):
        self._set_buttons_enabled(not self._is_deleted[self._current_row])

    def _set_strikeout_font(self, row, a0):
        for item in self._table_items[row]:
            f = item.font()
            f.setStrikeOut(a0)
            item.setFont(f)

    @QtCore.Slot()
    def _add_row_dialog(self):
        item_filter_list = self._get_item_filter_list(include_current_row=True)
        dialog = TableRowEditor(header=self._header, item_filter_list=item_filter_list,
                                **self._row_editor_kwargs, parent=self)
        dialog.data_collected.connect(self._add_row)
        dialog.exec()

    @QtCore.Slot(list)
    def _add_row(self, data_row: list[str]):
        row = []
        for data_item in data_row:
            row.append(QtWidgets.QTableWidgetItem(data_item))
        self._table_items.append(row)

        self._table.setRowCount(len(self._table_items))
        for j, item in enumerate(row):
            self._table.setItem(len(self._table_items) - 1, j, item)

        self._is_deleted += [False]

    @QtCore.Slot()
    def _delete_row(self):
        current_row = self._current_row

        if current_row >= len(self._old_data) or not self._remember_deleted:
            self._table_items.pop(current_row)
            self._table.removeRow(current_row)

            self._is_deleted.pop(current_row)
            current_row -= 1
        else:
            self._is_deleted[current_row] = True
            self._set_strikeout_font(current_row, True)

        self._table.selectionModel().clearSelection()
        self._set_buttons_enabled(False)

        if current_row + 1 < len(self._table_items):
            current_row += 1
            self._table.setCurrentCell(current_row, 0)
            self._table_items[current_row][0].tableWidget().setFocus()

    @QtCore.Slot()
    def _edit_row_dialog(self):
        item_filter_list = self._get_item_filter_list(include_current_row=False)
        dialog = TableRowEditor(header=self._header, data=self._new_data[self._current_row],
                                item_filter_list=item_filter_list, **self._row_editor_kwargs, parent=self)
        dialog.data_collected.connect(partial(self._edit_row, self._current_row))
        dialog.exec()

    @QtCore.Slot(list)
    def _edit_row(self, row_index: int, row_data: list[str]):
        for j, item in enumerate(self._table_items[row_index]):
            item.setText(row_data[j])

    @QtCore.Slot()
    def _reset(self):
        self._create_table_items()
        self._set_table_items()

        self._table.selectionModel().clearSelection()
        self._set_buttons_enabled(False)
        self.setFocus()

    @QtCore.Slot()
    def collect(self):
        self.data_collected.emit(self._new_data, self._is_deleted)



from qtpy import QtCore, QtWidgets

from ..config import config
from ..io.inspection_data import InspectionData, REDSHIFT_FILL_VALUE
from ..utils.widgets import AbstractWidget, MyQTextEdit


class InspectionResults(AbstractWidget):
    data_collected = QtCore.Signal(float, str, dict)
    edit_button_clicked = QtCore.Signal()

    def __init__(self, cfg: config.InspectionResults, parent=None):

        self.cfg = cfg
        self._saved_redshift: float | None = None

        self._redshift_widget: QtWidgets.QLabel | None = None
        self._spacer: QtWidgets.QWidget | None = None
        self._clear_redshift: QtWidgets.QPushButton | None = None
        self._separator: QtWidgets.QFrame | None = None

        self._checkbox_widgets: dict[str, QtWidgets.QCheckBox] | None = None
        self._comments_widget: MyQTextEdit | None = None

        self._edit_fields: QtWidgets.QPushButton | None = None

        super().__init__(parent=parent)
        self.setEnabled(False)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)

        # add shortcuts
        QtWidgets.QShortcut("E", self, self._comments_widget.setFocus)

    def _create_checkbox_widgets(self, review: InspectionData | None = None):
        if self._checkbox_widgets is not None:
            for w in self._checkbox_widgets.values():
                w.deleteLater()

        flags = self.cfg.default_flags if review is None else review.flag_columns

        checkbox_widgets = {}
        for i, flag_name in enumerate(flags):
            # TODO: overwrite the keyPressEvent method of QtWidgets.QCheckBox to add shortcuts
            checkbox_widget = QtWidgets.QCheckBox(flag_name, self)
            checkbox_widget.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
            checkbox_widgets[flag_name] = checkbox_widget

            if i < 9:
                checkbox_widget.setShortcut(str(i + 1))

        self._checkbox_widgets = checkbox_widgets

    def init_ui(self):
        self._redshift_widget = QtWidgets.QLabel("Redshift: --", self)
        self._spacer = QtWidgets.QWidget(self)
        self._spacer.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self._clear_redshift = QtWidgets.QPushButton("Clear", self)

        self._separator = QtWidgets.QFrame(self)
        self._separator.setFrameShape(QtWidgets.QFrame.HLine)

        self._create_checkbox_widgets()
        self._comments_widget = MyQTextEdit(self)
        self._comments_widget.setPlaceholderText('Comment')
        self._comments_widget.setMinimumWidth(90)

        self._edit_fields = QtWidgets.QPushButton("Edit...", self)
        self._edit_fields.clicked.connect(self.edit_button_clicked.emit)

        self._clear_redshift.clicked.connect(self.clear_redshift_value)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._redshift_widget)
        sub_layout.addWidget(self._spacer)
        sub_layout.addWidget(self._clear_redshift)
        self.layout().addLayout(sub_layout)

        self.layout().addWidget(self._separator)
        self._separator.setVisible(False)

        for i, widget in enumerate(self._checkbox_widgets.values()):
            self.layout().addWidget(widget)

        self.layout().addWidget(self._comments_widget)

        self.layout().addWidget(self._edit_fields)

    @QtCore.Slot(InspectionData)
    def load_project(self, review: InspectionData):
        self.setEnabled(True)

        self._create_checkbox_widgets(review=review)
        self.repopulate()

    @QtCore.Slot(float)
    def set_redshift_value(self, redshift: float):
        if redshift != REDSHIFT_FILL_VALUE:
            self._redshift_widget.setText(f"Redshift: {redshift:.4f}")
        else:
            self._redshift_widget.setText(f"Redshift: --")
        self._saved_redshift = redshift

    @QtCore.Slot()
    def clear_redshift_value(self):
        self.set_redshift_value(REDSHIFT_FILL_VALUE)

    @QtCore.Slot(int, InspectionData)
    def load_object(self, j: int, review: InspectionData):
        self.set_redshift_value(review.get_value(j, 'z_sviz'))
        self._comments_widget.setText(review.get_value(j, 'comment'))
        for cname, widget in self._checkbox_widgets.items():
            widget.setChecked(review.get_value(j, cname))

    @QtCore.Slot(int, InspectionData)
    def update_inspection_fields(self, j: int, review: InspectionData):
        self._create_checkbox_widgets(review=review)
        self.repopulate()

        self.load_object(j, review)  # not the "cleanest" solution

    @QtCore.Slot()
    def collect(self):
        redshift = self._saved_redshift
        comment = self._comments_widget.toPlainText()
        checkboxes = {cname: widget.isChecked() for cname, widget in self._checkbox_widgets.items()}

        self.data_collected.emit(redshift, comment, checkboxes)



from astropy.convolution import convolve_fft, Gaussian2DKernel
from astropy.visualization import ZScaleInterval
from astropy.wcs import WCS, FITSFixedWarning
import numpy as np
import pyqtgraph as pg
from qtpy import QtCore, QtGui

from dataclasses import dataclass
import logging
import warnings

from ..config import data_widgets
from ..io.catalog import Catalog
from ..utils.qt_tools import get_qtransform_from_wcs
from ..utils.widgets import ColorBar

from .ViewerElement import ViewerElement, LinkableItem


__all__ = [
    "Image2D"
]

logger = logging.getLogger(__name__)


@dataclass
class Image2DLevels:
    min: float = 0
    max: float = 1


class Image2D(ViewerElement):
    allowed_data_types: tuple[type] = (np.ndarray,)
    allowed_cbar_lims: tuple[str] = ('minmax', 'zscale', 'user')

    def __init__(self, cfg: data_widgets.Image, **kwargs):
        self.cfg = cfg

        self.image_item: pg.ImageItem | None = None

        self._cmap = pg.colormap.get('viridis')
        self.cbar: ColorBar | None = None
        self._default_levels: Image2DLevels | None = None

        super().__init__(cfg=cfg, **kwargs)

    def init_ui(self):
        super().init_ui()

        # lock the aspect ratio of the container
        self.container.setAspectLocked(True)

        # create an image item
        self.image_item = pg.ImageItem()
        self.image_item.setLookupTable(self._cmap.getLookupTable())
        self.image_item.setBorder('k')  # add a border to the image

        # create a color bar
        self.cbar = ColorBar(imageItem=self.image_item, showHistogram=True, histHeightPercentile=99.0)
        self.cbar.setVisible(self.cfg.color_bar.visible)
        self.cbar.axisItem.setVisible(False)

    def populate(self):
        super().populate()

        # add the color bar to the layout
        self.graphics_layout.addItem(self.cbar, 0, 1)

    def init_view(self):
        super().init_view()
        self._default_levels = Image2DLevels()

    def add_content(self):
        self.image_item.setImage(self.data, autoLevels=False)
        self.register_item(self.image_item)

        # add axes of symmetry to the image
        if self.cfg.central_axes.x or self.cfg.central_axes.y:
            pen = 'w'
            x, y = self.data.shape[1], self.data.shape[0]

            if self.cfg.central_axes.x:
                self.register_item((pg.PlotCurveItem([0, x], [y / 2, y / 2], pen=pen)))
            if self.cfg.central_axes.y:
                self.register_item((pg.PlotCurveItem([x / 2, x / 2], [0, y], pen=pen)))

        # add a crosshair to the image
        if self.cfg.central_crosshair:
            pen = pg.mkPen('w', width=1, style=QtCore.Qt.DashLine)
            x0, y0 = self.data.shape[1] // 2, self.data.shape[0] // 2
            dx, dy = 0.15 * x0, 0.15 * y0

            self.register_item(pg.PlotCurveItem([0, x0 - dx], [y0, y0], pen=pen))
            self.register_item(pg.PlotCurveItem([x0, x0], [0, y0 - dy], pen=pen))

    @property
    def has_defined_levels(self) -> bool:
        return np.any(np.isfinite(self.data)) and np.any(np.nonzero(self.data))

    def setup_view(self, cat_entry: Catalog | None):
        self.set_qtransform(cat_entry)
        self.set_default_range((0., float(self.data.shape[1])), (0., float(self.data.shape[0])),
                               apply_qtransform=True)

        # compute default image levels
        if self.has_defined_levels:
            limits_cfg = self.cfg.color_bar.limits
            if limits_cfg.type not in self.allowed_cbar_lims:
                logger.error(f'Unknown type of colorbar limits: {limits_cfg}.'
                             f'Supported types: {self.allowed_cbar_lims}')
            else:
                if limits_cfg.type == 'minmax':
                    l1, l2 = np.nanmin(self.data), np.nanmax(self.data)
                elif limits_cfg.type == 'zscale':
                    l1, l2 = ZScaleInterval().get_limits(self.data)
                else:
                    l1 = limits_cfg.min if limits_cfg.min is not None else np.nanmin(self.data)
                    l2 = limits_cfg.max if limits_cfg.max is not None else np.nanmax(self.data)

                self.set_default_levels((l1, l2))

        super().setup_view(cat_entry)

    def set_qtransform(self, cat_entry: Catalog | None):
        qtransform = QtGui.QTransform()

        # create a transformation matrix from metadata
        if self.cfg.wcs_transform and self.meta is not None:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FITSFixedWarning)
                w = WCS(self.meta)

            qtransform *= get_qtransform_from_wcs(w)

        # rotate the image
        rotation_angle = self.cfg.rotate
        if self.cfg.rotate == "auto":
            try:
                rotation_angle = float(cat_entry.get_col("PA"))
            except (TypeError, KeyError) as e:
                logger.warning(e)
                rotation_angle = None
        if rotation_angle:
            qtransform = qtransform.rotate(rotation_angle)

        self._qtransform = qtransform

    def set_default_levels(self, levels: tuple[float, float]):
        self._default_levels.min = levels[0]
        self._default_levels.max = levels[1]

    def set_levels(self, levels: tuple[float, float]):
        self.cbar.setLevels(levels)
        self.cbar._updateHistogram()  # the histogram is calculated using the current image levels

    def apply_qtransform(self, **kwargs):
        super().apply_qtransform(**kwargs)
        self.container.setAspectLocked(lock=True, ratio=self._qtransform.m22() / self._qtransform.m11())

    def smooth(self, sigma: float):
        if sigma > 0:
            # create a smoothing kernel
            gauss_kernel = Gaussian2DKernel(sigma)

            # FFT algorithm is faster for large arrays (n > 500), which is a typical case for astronomy images
            smoothed_data = convolve_fft(self.data, gauss_kernel, preserve_nan=True)
        else:
            smoothed_data = self.data

        self.image_item.setImage(smoothed_data, autoLevels=False)

    def reset_levels(self):
        self.set_levels((self._default_levels.min, self._default_levels.max))

    @QtCore.Slot(object)
    def reset_view(self, widget_links: dict | None = None):
        super().reset_view(widget_links=widget_links)

        if widget_links is None:
            widget_links = dict()

        if self.title not in widget_links.get(LinkableItem.COLORBAR, dict()):
            self.reset_levels()

    def _enter_zen_mode(self):
        super()._enter_zen_mode()
        self.cbar.setVisible(False)

    def _update_visibility(self):
        super()._update_visibility()
        self.cbar.setVisible(self.cfg.color_bar.visible)



from functools import wraps
import logging

from qtpy import QtWidgets


class QLogHandler(logging.Handler):
    def emit(self, record):
        msg = record.msg
        if isinstance(msg, Exception):
            msg = str(msg)
            if msg.startswith("'") and msg.endswith("'"):
                msg = msg[1:-1]

        level = record.levelno
        if level <= logging.INFO:
            return
        LogMessageBox(level, msg, parent=QtWidgets.QApplication.focusWidget())


class LogMessageBox(QtWidgets.QMessageBox):
    PACKAGE = __package__.split('.')[0].capitalize()  # the name of the top-level package

    LOGGING_LEVELS = {
        logging.INFO: QtWidgets.QMessageBox.Information,
        logging.WARNING: QtWidgets.QMessageBox.Warning,
        logging.ERROR: QtWidgets.QMessageBox.Warning,
        logging.CRITICAL: QtWidgets.QMessageBox.Critical
    }

    def __init__(self, level: int, message: str, parent=None):
        super().__init__(self.LOGGING_LEVELS[level], '{} Message'.format(self.PACKAGE), message, parent=parent)
        self.show()


def qlog(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        root_logger = logging.getLogger()
        handler = QLogHandler()
        root_logger.addHandler(handler)

        res = func(*args, **kwargs)

        root_logger.removeHandler(handler)

        return res

    return wrapper



import pyqtgraph as pg

from ..config import Appearance
from .qdarktheme import setup_theme as setup_qtheme


__all__ = [
    "setup_appearance"
]


def setup_theme(theme: str):
    setup_qtheme(theme)

    if theme == 'dark':
        pg.setConfigOption('background', "#1d2023")
        pg.setConfigOption('foreground', '#eff0f1')
    else:
        pg.setConfigOption('background', 'w')
        pg.setConfigOption('foreground', 'k')


def setup_appearance(cfg: Appearance, update_theme: bool = True):
    pg.setConfigOption('antialias', cfg.antialiasing)

    if update_theme:
        setup_theme(theme=cfg.theme)



from astropy.table import Table
from qtpy import QtWidgets, QtCore

import logging
import pathlib

from ..io.catalog import Catalog
from ..io.inspection_data import InspectionData
from ..utils.widgets import AbstractWidget


logger = logging.getLogger(__name__)


class Subsets(AbstractWidget):
    inspect_button_clicked = QtCore.Signal()
    pause_inspecting_button_clicked = QtCore.Signal()
    stop_inspecting_button_clicked = QtCore.Signal()

    def __init__(self, parent=None):
        self._inspect_subset: QtWidgets.QPushButton | None = None
        self._pause_inspecting: QtWidgets.QPushButton | None = None
        self._stop_inspecting: QtWidgets.QPushButton | None = None
        self._subset_info: QtWidgets.QLabel | None = None

        self._subset_cat: Catalog | None = None
        self._subset_name: str | None = None

        super().__init__(parent=parent)
        self.setEnabled(False)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)

    def init_ui(self):
        self._inspect_subset = QtWidgets.QPushButton('Inspect Subset...', parent=self)
        self._pause_inspecting = QtWidgets.QPushButton('Pause', parent=self)
        self._stop_inspecting = QtWidgets.QPushButton('Stop', parent=self)

        self._inspect_subset.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        for w in (self._pause_inspecting, self._stop_inspecting):
            w.setFixedWidth(90)

        self._pause_inspecting.setEnabled(False)
        self._stop_inspecting.setEnabled(False)

        self._subset_info = QtWidgets.QLabel(self)
        self._subset_info.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        self._subset_info.setVisible(False)

        self._inspect_subset.clicked.connect(self.inspect_button_clicked.emit)
        self._pause_inspecting.clicked.connect(self.pause_inspecting_button_clicked.emit)
        self._stop_inspecting.clicked.connect(self.stop_inspecting_button_clicked.emit)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._inspect_subset)
        self.layout().addLayout(sub_layout)

        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._pause_inspecting)
        sub_layout.addWidget(self._stop_inspecting)
        self.layout().addLayout(sub_layout)

        self.layout().addWidget(self._subset_info)

    @QtCore.Slot()
    def load_project(self):
        self.setEnabled(True)

    @QtCore.Slot(str, object)
    def load_subset(self, subset_path: str, subset: Table):
        self._subset_cat = subset
        self._subset_name = pathlib.Path(subset_path).name

        self._pause_inspecting.setEnabled(True)
        self._stop_inspecting.setEnabled(True)

        self.set_subset_info()
        self._subset_info.setVisible(True)

    def set_subset_info(self, obj_str: str | None = None):
        if obj_str is None:
            obj_str = '--'
        self._subset_info.setText(f'Subset: {self._subset_name}\nObject: {obj_str}/{len(self._subset_cat)}')

    @QtCore.Slot(int, InspectionData)
    def load_object(self, j: int, review: InspectionData):
        if self._subset_cat:
            subset_entry = self._subset_cat.get_cat_entry(review.get_id(j, full=True), ignore_missing=True)
            if subset_entry is None:
                self.set_subset_info()
            else:
                j_subset = subset_entry.get_col('__index__')
                self.set_subset_info(str(j_subset + 1))

    @QtCore.Slot(bool)
    def pause_subset_inspection(self, is_paused: bool):
        self._pause_inspecting.setText("Resume" if is_paused else "Pause")

    @QtCore.Slot()
    def stop_subset_inspection(self):
        self._subset_cat = None
        self._subset_name = None
        self._pause_inspecting.setText('Pause')

        self._pause_inspecting.setEnabled(False)
        self._stop_inspecting.setEnabled(False)
        self._subset_info.setVisible(False)



from qtpy import QtCore

import logging
import pathlib
import time

from ..config import config
from ..io.catalog import Catalog
from ..io.inspection_data import InspectionData
from ..io.viewer_data import ViewerData, LocalPath

from .ViewerElement import ViewerElement


__all__ = [
    "ViewerDataLoader"
]

logger = logging.getLogger(__name__)


class ViewerDataLoader(QtCore.QThread):
    data_loaded = QtCore.Signal(object, object, object)

    def __init__(self, widgets: dict[str, ViewerElement], j: int, review: InspectionData, viewer_data: ViewerData,
                 data_sources: config.DataSources, cat_entry: Catalog | None, t_grace=0.1):
        super().__init__(parent=None)

        self.widgets: dict[str, ViewerElement] = widgets
        
        self.j: int = j
        self.review: InspectionData = review
        self.data_sources: config.DataSources = data_sources
        self.cat_entry: Catalog | None = cat_entry
        self.viewer_data = viewer_data

        self.t_grace = t_grace
        
        self._runs = True

    def run(self):
        i = 0
        n = 30
        dt = self.t_grace / n
        while self._runs and i < n:
            time.sleep(dt)
            i += 1

        if i < n:
            return

        widgets = list(self.widgets.values())

        i = 0
        n = len(widgets)
        while self._runs and i < n:
            w0 = widgets[i]
            res = self._load_data(w0)
            if res is None:
                res = (None, None, None)
            self.data_loaded.connect(w0.set_data)
            self.data_loaded.emit(*res)
            self.data_loaded.disconnect(w0.set_data)
            i += 1

    @QtCore.Slot()
    def abort(self):
        self._runs = False
    
    def _load_data(self, w0: ViewerElement):
        if w0.data is not None:
            self._free_resources(w0)
        
        data_path = self._get_data_path(w0)
        if data_path is None:
            return None

        try:
            data_path.resolve(self.review.get_id(self.j), self.cat_entry)
        except Exception as e:
            logger.error(f"Failed to resolve the filename: {e} (widget: {w0.title})")
            return None

        try:
            data_path.validate()
        except FileNotFoundError:
            logger.error(f"`{w0.title}` not found (filename: {data_path.name})")
            return None
        except Exception as e:
            logger.error(f"{e} (widget: {w0.title})")
            return None

        loader_params = self._get_loader_params(w0)
        if loader_params is None:
            return None

        data, meta = self.viewer_data.load(str(data_path),
                                           loader=w0.cfg.data.loader,
                                           allowed_dtypes=w0.allowed_data_types,
                                           **loader_params)

        return data, meta, data_path

    def _free_resources(self, w0: ViewerElement):
        if not w0.cfg.data.source:
            self.viewer_data.close(str(w0.data_path))
            return

        self.viewer_data.reopen(str(w0.data_path))

    def _get_data_path(self, w0: ViewerElement) -> LocalPath | None:
        if not w0.cfg.data.source:
            if not w0.cfg.data.filename:
                logger.error(f"Filename not specified (widget: {w0.title})")
                return
            return LocalPath(str(pathlib.Path(self.data_sources.dir) / w0.cfg.data.filename))

        # for now assume that a non-empty data source == image
        image = self.data_sources.images.get(w0.cfg.data.source)
        if not image:
            logger.error(f"Shared image not found (label: {w0.cfg.data.source}, widget: {w0.title})")
            return

        return LocalPath(image.filename)

    def _get_loader_params(self, w0: ViewerElement) -> dict | None:
        params = w0.cfg.data.loader_params

        if not w0.cfg.data.source:
            return params

        if self.cat_entry is None:
            logger.error(f"Failed to create an image cutout: Catalog entry not loaded (widget: {w0.title})")
            return None

        image = self.data_sources.images.get(w0.cfg.data.source)
        if not image:
            logger.error(f"Shared image not found (label: {w0.cfg.data.source}, widget: {w0.title})")
            return None

        wcs_source = image.wcs_source if image.wcs_source else image.filename
        cutout_params = self._get_cutout_params(wcs_source)
        if cutout_params is None:
            return None

        params.update(cutout_params)

        return params

    def _get_cutout_params(self, wcs_source: str) -> dict | None:
        params = dict(create_cutout=True)

        try:
            ra = self.cat_entry.get_col('ra')
            dec = self.cat_entry.get_col('dec')
        except KeyError as e:
            logger.error(e)
            return None

        _, meta = self.viewer_data.load(wcs_source, lazy=True, create_wcs=True)
        if meta is None:
            return

        try:
            x0, y0 = meta.wcs.all_world2pix(ra, dec, 0)
        except Exception as e:
            logger.error(f"Calculation of pixel coordinates failed: {e} (image: {wcs_source})")
            return None
        params.update(x0=x0, y0=y0)

        return params


from dataclasses import dataclass, field

from ..utils.params import Params


@dataclass
class Cache(Params):
    last_inspection_file: str | None = None
    last_object_index: int | None = None
    last_subset_file: str | None = None
    visible_columns: list[str] | None = None
    dock_layout: dict | None = None



import numpy as np
from astropy.table import Table
from astropy.units import Quantity
import pyqtgraph as pg
from scipy.ndimage import gaussian_filter1d

import logging

from ..config import data_widgets
from ..io.catalog import Catalog

from .ViewerElement import ViewerElement


__all__ = ["Plot1D"]

logger = logging.getLogger(__name__)


class Plot1D(ViewerElement):
    allowed_data_types = (Table,)

    def __init__(self, cfg: data_widgets.Plot1D, **kwargs):
        self.cfg = cfg

        self.plot_data_items: dict[str, pg.PlotDataItem] = {}

        super().__init__(cfg=cfg, **kwargs)

    def get_plot_data(self, cname: str) -> Quantity | None:
        try:
            plot_data = self.data[cname].quantity
        except KeyError:
            logger.warning(f"Column not found: {cname} (widget: {self.title})")
            return None

        return Quantity(plot_data)  # return a copy to prevent any modifications to self.data

    def add_content(self):
        default_pen = pg.getConfigOption('foreground')
        for label, line_plot in self.cfg.plots.items():
            x_data, y_data = self.get_plot_data(line_plot.x), self.get_plot_data(line_plot.y)
            if x_data is None or y_data is None:
                continue

            # if axis labels are not set, adopt the plot labels
            if not self._axes.x.label:
                self._axes.x.label = line_plot.x
            if not self._axes.y.label:
                self._axes.y.label = line_plot.y

            # if axis units are not set, adopt the units of the data
            if not self._axes.x.unit:
                self._axes.x.unit = x_data.unit
            if not self._axes.y.unit:
                self._axes.y.unit = y_data.unit

            # apply scaling and unit conversion
            x_data, y_data = self.apply_xdata_transform(x_data), self.apply_ydata_transform(y_data)

            # add a legend to the plot
            pen = default_pen if line_plot.color is None else line_plot.color
            name = label if not line_plot.hide_label else None
            if name and self.container.legend is None:
                self.container.addLegend(verSpacing=-10, pen=default_pen)

            # create and register a plot data item
            plot_data_item = pg.PlotDataItem(x=x_data, y=y_data, pen=pen, name=name)
            self.plot_data_items[label] = plot_data_item
            self.register_item(plot_data_item)

    @staticmethod
    def calc_axis_lims(lims_current: tuple[float, float] | None, plot_data: np.ndarray):
        plot_data = plot_data[~np.isinf(plot_data)]
        if all(np.isnan(plot_data)):
            return lims_current
        lims_new = (np.nanmin(plot_data), np.nanmax(plot_data))

        if lims_current is not None:
            lims_new = (min(lims_current[0], lims_new[0]), max(lims_current[1], lims_new[1]))

        return lims_new

    def setup_view(self, cat_entry: Catalog | None):
        xlim, ylim = None, None
        for label, plot_data_item in self.plot_data_items.items():
            x_data, y_data = plot_data_item.getData()
            xlim = self.calc_axis_lims(xlim, x_data)
            ylim = self.calc_axis_lims(ylim, y_data)

        self.set_default_range(xlim, ylim)
        self.set_content_padding(ypad=0.05)

        super().setup_view(cat_entry)

    def smooth(self, sigma: float):
        for label, plot_data_item in self.plot_data_items.items():
            x_data, _ = plot_data_item.getData()
            y_data = self.get_plot_data(self.cfg.plots[label].y)
            y_data = self.apply_ydata_transform(y_data)

            y_data_smoothed = gaussian_filter1d(y_data, sigma) if sigma > 0 else y_data
            plot_data_item.setData(x=x_data, y=y_data_smoothed)

    def clear_content(self):
        # TODO: submit issue to the pyqtgraph repo
        # if self.container.legend:
        #     self.container.vb.removeItem(self.container.legend)  # removing from the ViewBox, not PlotItem

        super().clear_content()



from astropy.table import Table
import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import logging
import pathlib


logger = logging.getLogger(__name__)

REDSHIFT_FILL_VALUE = -1.0


class WriterBase(ABC):
    @abstractmethod
    def write(self, df: pd.DataFrame, filename: pathlib.Path):
        pass


class CSVWriter(WriterBase):
    def write(self, df: pd.DataFrame, filename: pathlib.Path):
        df.to_csv(filename)


class FITSWriter(WriterBase):
    def write(self, df: pd.DataFrame, filename: pathlib.Path):
        t: Table = Table.from_pandas(df.reset_index())
        t.write(filename, overwrite=True)


@dataclass
class InspectionData:
    df: pd.DataFrame
    default_columns: list[str] = field(default_factory=lambda: ['starred', 'z_sviz', 'comment'])

    @staticmethod
    def _add_default_columns(df: pd.DataFrame):
        # objects starred by the user
        if 'starred' not in df.columns:
            df['starred'] = False

        # redshifts saved by the user
        if 'z_sviz' not in df.columns:
            df['z_sviz'] = REDSHIFT_FILL_VALUE

        # column for user comments
        if 'comment' not in df.columns:
            df['comment'] = ''
        else:
            df['comment'] = df['comment'].fillna('')

        return df

    @classmethod
    def create(cls, *args, flags: list[str] | None = None):
        """ Create a new instance of the InspectionData class with a Pandas dataframe containing:
              - a column of IDs;
              - a column for comments;
              - one column per each user-defined flag.
        @param flags: the list of user-defined flags
        @return: an instance of the InspectionData class
        """

        try:
            if len(args) == 1:
                index = pd.Index(args[0], name='id')
            else:
                index = pd.MultiIndex.from_arrays(args, names=('id',) + tuple(f'id{i + 1}' for i in range(1, len(args))))
        except TypeError as e:
            logger.error(f'Failed to create the inspection file: {e}')
            return None

        df = pd.DataFrame(index=index)
        df = cls._add_default_columns(df)

        review = cls(df=df)

        if flags is not None:
            for cname in flags:
                review.add_flag_column(cname)

        return review

    @classmethod
    def read(cls, filename: str | pathlib.Path):
        """ Read an existing inspection file
        @param filename: the input filename
        @return: an instance of the InspectionData class
        """

        # TODO: validate the input
        df = pd.read_csv(filename, index_col='id')

        i = 1
        while f'id{i + 1}' in df.columns:
            df.set_index(f'id{i + 1}', append=True, inplace=True)
            i += 1

        df = cls._add_default_columns(df)

        review = cls(df=df)
        review.reorder_columns()

        return review

    def write(self, filename: pathlib.Path, fmt: str = 'csv'):
        """ Write inspection data to the output file.
        @param filename: the output filename
        @param fmt: the output format
        @return: None
        """

        writers: dict[str, type[WriterBase]] = {
            'csv': CSVWriter,
            'fits': FITSWriter
        }

        if writers.get(fmt):
            writers[fmt]().write(self.df, filename)
        else:
            logger.error(f"Unknown output format: {fmt}")

    @property
    def n_objects(self) -> int | None:
        """
        @return: the total number of objects under inspection.
        """
        return len(self.df)

    @property
    def ids(self):
        """ Primary IDs.
        """
        return self.df.index.get_level_values(0)

    @property
    def ids_full(self):
        """ All available IDs.
        """
        return self.df.index.values

    @property
    def ids_are_int(self) -> bool:
        return pd.api.types.is_integer_dtype(self.ids)

    @property
    def indices(self) -> list[str]:
        return self.df.index.names

    @property
    def user_defined_columns(self) -> list[str]:
        return [cname for cname in self.df.columns if cname not in self.default_columns]

    @property
    def flag_columns(self) -> list[str]:
        return [cname for cname in self.user_defined_columns if pd.api.types.is_bool_dtype(self.df[cname])]

    def add_flag_column(self, column_name: str):
        self.df[column_name] = False

    def reorder_columns(self):
        self.df = self.df[self.default_columns + self.user_defined_columns]

    def rename_column(self, old_name: str, new_name: str):
        if old_name not in self.user_defined_columns:
            logger.error(f"Failed to rename a column: Column not found (column: {old_name})")
            return
        self.df.rename(columns={old_name: new_name}, inplace=True)

    def delete_column(self, column_name: str):
        if column_name not in self.user_defined_columns:
            logger.error(f"Failed to delete a column: Column not found (column: {column_name})")
            return
        self.df.drop(column_name, axis=1, inplace=True)

    def to_list(self):
        return self.df.values.tolist()

    def has_data(self, column_name: str) -> bool:
        if column_name in self.flag_columns or column_name == 'starred':
            return self.df[column_name].sum() > 0
        else:
            logger.warning(f"Cannot determine if a column has data or not (column: {column_name})")
            return True

    def get_value(self, j: int, cname: str):
        return self.df.iat[j, self.df.columns.get_loc(cname)]

    def get_id(self, j: int, full=False) -> str | int | tuple | None:
        try:
            obj_id = self.ids[j] if not full else self.ids_full[j]
        except (TypeError, IndexError):
            return

        if self.ids_are_int and not isinstance(obj_id, tuple):
            # convert int64 to int
            obj_id = int(obj_id)

        return obj_id

    def get_id_loc(self, obj_id: str | int) -> int:
        if self.ids_are_int:
            obj_id = int(obj_id)

        j = self.df.index.get_loc(obj_id)
        if isinstance(j, slice):
            j = j.start  # use the first available secondary ID
        elif isinstance(j, np.ndarray):
            j = np.argmax(j)

        return int(j)

    def update_value(self, j: int, cname: str, value):
        self.df.iat[j, self.df.columns.get_loc(cname)] = value

    def validate_id(self, obj_id: str | int) -> bool:
        if self.ids_are_int:
            try:
                obj_id = int(obj_id)
            except ValueError:
                logger.error(f'Invalid ID: {obj_id}')
                return False

        if obj_id not in self.ids:
            logger.error(f'ID `{obj_id}` not found')
            return False

        return True

    def validate_index(self, index: int) -> bool:

        if not 0 < index <= self.n_objects:
            logger.error(f'Index `{index}` out of range')
            return False

        return True



from qtpy import QtWidgets, QtCore

import logging
import pathlib

from ..config import config
from ..io.catalog import Catalog, cat_browser
from ..io.data_dir import get_ids_from_dir
from ..io.viewer_data import data_browser
from ..utils.logs import qlog
from ..utils.widgets import FileBrowser

logger = logging.getLogger(__name__)


class NewFile(QtWidgets.QDialog):
    output_path_selected = QtCore.Signal(pathlib.Path)
    catalogue_changed = QtCore.Signal(object)

    def __init__(self, cfg: config.Config, parent=None):
        self.cfg = cfg

        self._browsers: dict[str, FileBrowser] | None = None
        self._cat_factory_group: QtWidgets.QButtonGroup | None = None
        self._create_cat_radio_button: QtWidgets.QRadioButton | None = None
        self._load_cat_radio_button: QtWidgets.QRadioButton | None = None
        self._separator: QtWidgets.QFrame | None = None
        self._filter_check_box: QtWidgets.QCheckBox | None = None
        self._id_pattern_label: QtWidgets.QLabel | None = None
        self._id_pattern: QtWidgets.QLineEdit | None = None
        self._spacer: QtWidgets.QWidget | None = None
        self._button_box: QtWidgets.QDialogButtonBox | None = None

        super().__init__(parent)
        self.setWindowTitle("Create a New Inspection File")

        self.init_ui()
        self.set_layout()
        self.populate()

        self.setFixedWidth(self.layout().sizeHint().width())

    def init_ui(self):
        width = 135
        self._browsers = {
            'output': FileBrowser(filename_extensions='CSV Files (*.csv)',
                                  mode=FileBrowser.SaveFile, default_path=pathlib.Path().resolve() / 'Untitled.csv',
                                  title='Output File:', title_width=width, parent=self),
            'data': data_browser(self.cfg.data.dir, title='Data Source:', title_width=width, parent=self),
            'cat': cat_browser(self.cfg.catalogue.filename, title='Catalogue:', title_width=width, parent=self)
        }

        # add radio buttons for choosing between creating a new catalogue and loading an existing one
        self._cat_factory_group = QtWidgets.QButtonGroup(parent=self)
        self._cat_factory_group.buttonToggled.connect(self.update_cat_factory_state)

        self._create_cat_radio_button = QtWidgets.QRadioButton('Create a new catalogue', self)
        self._cat_factory_group.addButton(self._create_cat_radio_button)

        self._load_cat_radio_button = QtWidgets.QRadioButton('Load an existing catalogue', self)
        self._cat_factory_group.addButton(self._load_cat_radio_button)

        # create a horizontal separator
        self._separator = QtWidgets.QFrame(parent=self)
        self._separator.setFrameShape(QtWidgets.QFrame.HLine)

        # add a checkbox for specifying the catalogue loader mode
        self._filter_check_box = QtWidgets.QCheckBox(
            'Filter the catalogue using a list of IDs extracted from the data directory', self)

        self._filter_check_box.stateChanged.connect(self.update_id_pattern_widget_state)

        # add a line edit for specifying the ID pattern that will be used to parse the filenames in the data directory
        self._id_pattern_label = QtWidgets.QLabel('ID pattern:', self)
        self._id_pattern_label.setToolTip('A regular expression used to match the object IDs during the scan of '
                                          'the data directory. If more than\none ID is matched to the pattern for a '
                                          'given filename, only the longest match will be returned.')
        self._id_pattern_label.setFixedWidth(120)
        self._id_pattern_label.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)

        self._id_pattern = QtWidgets.QLineEdit(self.cfg.data.id_pattern, self)

        self._spacer = QtWidgets.QWidget(self)
        self._spacer.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        # add OK/Cancel buttons
        self._button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
                                                      self)
        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)

        # set the state of radio buttons and checkboxes
        self._filter_check_box.setChecked(False)
        self._load_cat_radio_button.setChecked(True)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setSpacing(20)
        self.setFixedHeight(450)

    def populate(self):
        # add a file browser for specifying the output file
        self.layout().addWidget(self._browsers['output'])

        # add a file browser for specifying the data directory
        self.layout().addWidget(self._browsers['data'])

        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._load_cat_radio_button)
        sub_layout.addWidget(self._create_cat_radio_button)
        self.layout().addLayout(sub_layout)

        self.layout().addWidget(self._separator)

        # add a file browser for specifying the catalogue
        self.layout().addWidget(self._browsers['cat'])

        self.layout().addWidget(self._filter_check_box)

        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._id_pattern_label)
        sub_layout.addWidget(self._id_pattern)
        self.layout().addLayout(sub_layout)

        self.layout().addWidget(self._spacer)
        self.layout().addWidget(self._button_box)

    @qlog
    def validate(self) -> bool:
        for b in self._browsers.values():
            if not b.isHidden() and (not b.is_filled(verbose=True) or not b.exists(verbose=True)):
                return False
        return True

    @qlog
    def get_catalogue(self):
        data_dir = self._browsers['data'].path
        kwargs = dict(
            id_pattern=self._id_pattern.text(),
            recursive=True
        )

        # create a new catalogue if necessary
        if self._browsers['cat'].isHidden():
            ids = get_ids_from_dir(data_dir, **kwargs)
            if ids is None:
                return
            return Catalog.create(ids)

        # otherwise, load an existing catalogue
        if not self._filter_check_box.isChecked():
            data_dir = None
        return Catalog.read(self._browsers['cat'].path, translate=self.cfg.catalogue.translate, data_dir=data_dir,
                            **kwargs)

    def accept(self):
        if not self.validate():
            return

        cat = self.get_catalogue()
        if cat is None:
            return

        self.catalogue_changed.emit(cat)
        self.output_path_selected.emit(pathlib.Path(self._browsers['output'].path))

        # update the user configuration file
        self.cfg.data.dir = self._browsers['data'].path
        if not self._browsers['cat'].isHidden():
            self.cfg.catalogue.filename = self._browsers['cat'].path
        else:
            self.cfg.catalogue.filename = None
        if not self._id_pattern.isHidden():
            self.cfg.data.id_pattern = self._id_pattern.text()
        self.cfg.save()

        super().accept()

    def update_cat_factory_state(self):
        if self._create_cat_radio_button.isChecked():
            self._browsers['cat'].setHidden(True)
            self._filter_check_box.setHidden(True)

            self._id_pattern.setHidden(False)
            self._id_pattern_label.setHidden(False)
        else:
            self._browsers['cat'].setHidden(False)
            self._filter_check_box.setHidden(False)

            self.update_id_pattern_widget_state()

    def update_id_pattern_widget_state(self):
        if self._filter_check_box.isChecked():
            self._id_pattern.setHidden(False)
            self._id_pattern_label.setHidden(False)
        else:
            self._id_pattern.setHidden(True)
            self._id_pattern_label.setHidden(True)



from astropy.table import Table, Row, MaskedColumn
import numpy as np

from dataclasses import dataclass, field
import logging

from .data_dir import get_ids_from_dir
from ..utils.widgets import FileBrowser

logger = logging.getLogger(__name__)


@dataclass
class Catalog:
    table: Table | Row
    indices: list[str] = field(default_factory=list)
    translate: dict[str, list[str]] = field(default_factory=dict)

    def __post_init__(self):
        if isinstance(self.table, Table):
            for idx in self.indices:
                self.table.add_index(idx)

    @classmethod
    def create(cls, ids):
        colnames = ['id']
        if isinstance(ids[0], tuple):
            colnames += list(f'id{i + 1}' for i in range(1, len(ids[0])))
            table_data = list(zip(*ids))
        else:
            table_data = [ids]

        table = Table(table_data, names=colnames)
        logger.info("Catalog created")
        return cls(table=table, indices=colnames)

    def _add_indices(self) -> bool:
        # check that the ID column is present in the catalogue
        try:
            id_col = self.get_col('id')
        except KeyError as e:
            logger.error(e)
            return False

        self.add_index(id_col.name)
        for i in range(2, 11):
            try:
                id_col = self.get_col(f'id{i}')
            except KeyError:
                break
            else:
                self.add_index(id_col.name)

        for idx in self.indices:
            if isinstance(self.get_col(idx), MaskedColumn) and np.sum(self.get_col(idx).mask):
                logger.error(f"Some IDs are missing (column: {idx})")
                return False

        return True

    @classmethod
    def read(cls, filename: str, translate: dict[str, list[str]] | None = None, data_dir=None, **kwargs):
        """Read the catalogue from a file.
        @param filename: the catalogue filename
        @param translate:
        @param data_dir:
        @return: the Catalog object
        """
        try:
            table = Table.read(filename)  # load the catalogue
        except (OSError, ValueError) as e:
            logger.error(f"Failed to load the catalogue: {e}")
            return None

        cat = cls(table=table, translate=translate)
        if not cat._add_indices():
            return None

        if data_dir is not None:
            ids = get_ids_from_dir(data_dir, **kwargs)
            if ids is None:
                return None

            # filter objects based on the list of IDs
            cat.table = cat.table[np.isin(cat.get_col('id'), ids, assume_unique=False)]

        if len(cat.table) == 0:
            logger.error("The processed catalogue is empty")
            return None

        logger.info(f"Catalog loaded (path: {filename})")
        return cat

    def update_translate(self, new_translate: dict[str, list[str]] | None):
        cat = Catalog(table=self.table)

        cat.translate = new_translate
        if not cat._add_indices():
            return None

        return cat

    def __len__(self):
        return len(self.table)

    @property
    def colnames(self) -> list:
        return self.table.colnames

    @property
    def extended_colnames(self) -> list:
        colnames = self.colnames

        colnames_extension = []
        for cname, cname_aliases in self.translate.items():
            for cname_alias in cname_aliases:
                if cname_alias in colnames:
                    colnames_extension.append(cname)
                    break
        colnames += colnames_extension

        return colnames

    @property
    def annotated_colnames(self) -> dict[str, str]:
        colnames = {cname: cname for cname in self.colnames}

        for cname, cname_aliases in self.translate.items():
            for cname_alias in cname_aliases:
                if cname_alias in self.colnames:
                    colnames[cname_alias] = f'{cname_alias} ({cname})'
                    break

        return colnames

    def add_index(self, idx):
        self.indices.append(idx)
        self.table.add_index(idx)

    def add_column(self, data, **kwargs):
        self.table.add_column(data, **kwargs)

    def _loc_full_base(self, t: Table, indices: list, obj_id: str | int | tuple):
        if isinstance(obj_id, str) or isinstance(obj_id, int):
            return t.loc[obj_id]  # standard .loc method

        elif isinstance(obj_id, tuple):
            if not obj_id:
                return t  # exit recursion

            if len(indices) != len(obj_id):
                raise KeyError

            if isinstance(t, Row):
                if t[indices[0]] == obj_id[0]:
                    return t
                else:
                    raise KeyError

            return self._loc_full_base(t.loc[indices[0], obj_id[0]], indices[1:], obj_id[1:])
        else:
            raise TypeError(f'Unknown object ID type: {type(obj_id)}')

    def _loc_full(self, obj_id: str | int | tuple):
        """ Retrieve rows by (multi)index.
        """
        if self.indices is None:
            raise ValueError(f'Cannot locate the object: no indices found in the table (ID: {obj_id})')
        return self._loc_full_base(self.table, self.indices, obj_id)

    def get_col(self, cname: str):
        if cname not in self.extended_colnames:
            if cname in self.translate:
                raise KeyError('`{}` column and its aliases ({}) not found'.format(cname, ", ".join(self.translate[cname])))
            else:
                raise KeyError('`{}` column not found'.format(cname))

        if cname in self.colnames:
            return self.table[cname]

        for cname_alias in self.translate[cname]:
            if cname_alias in self.colnames:
                return self.table[cname_alias]

    def get_cat_entry(self, obj_id: str | int | tuple, ignore_missing=False):
        try:
            cat_entry = self._loc_full(obj_id)
        except KeyError:
            if not ignore_missing:
                logger.error(f'Object not found in the catalogue (ID: {obj_id})')
            return None
        except (TypeError, ValueError) as e:
            logger.error(e)
            return None

        if isinstance(cat_entry, Table):
            logger.error(f'Object corresponds to multiple entries in the catalogue (ID: {obj_id})')
            return None

        return Catalog(cat_entry, indices=self.indices, translate=self.translate)


def cat_browser(default_path, **kwargs) -> FileBrowser:
    return FileBrowser(filename_extensions='FITS Files (*.fits)', mode=FileBrowser.OpenFile,
                       default_path=default_path, **kwargs)



from .._color import Color
from .._icon.svg import Svg
from ..qtpy.QtCore import QPoint, QRect, QRectF, QSize, Qt
from ..qtpy.QtGui import (
    QGuiApplication,
    QIcon,
    QIconEngine,
    QImage,
    QPainter,
    QPalette,
    QPixmap,
)
from ..qtpy.QtSvg import QSvgRenderer


class SvgIconEngine(QIconEngine):
    """A custom QIconEngine that can render an SVG buffer."""

    def __init__(self, svg: Svg) -> None:
        """Initialize icon engine."""
        super().__init__()
        self._svg = svg

    def paint(self, painter: QPainter, rect: QRect, mode: QIcon.Mode, state):
        """Paint the icon int ``rect`` using ``painter``."""
        palette = QGuiApplication.palette()

        if mode == QIcon.Mode.Disabled:
            rgba = palette.color(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text).getRgb()
            color = Color.from_rgba(*rgba)
        else:
            rgba = palette.text().color().getRgb()
            color = Color.from_rgba(*rgba)
        self._svg.colored(color)

        svg_byte = str(self._svg).encode("utf-8")
        renderer = QSvgRenderer(svg_byte)  # type: ignore
        renderer.render(painter, QRectF(rect))

    def clone(self):
        """Required to subclass abstract QIconEngine."""
        return SvgIconEngine(self._svg)

    def pixmap(self, size: QSize, mode: QIcon.Mode, state: QIcon.State):
        """Return the icon as a pixmap with requested size, mode, and state."""
        # Make size to square.
        min_size = min(size.width(), size.height())
        size.setHeight(min_size)
        size.setWidth(min_size)

        img = QImage(size, QImage.Format.Format_ARGB32)
        img.fill(Qt.GlobalColor.transparent)
        pixmap = QPixmap.fromImage(img, Qt.ImageConversionFlag.NoFormatConversion)
        size.width()
        self.paint(QPainter(pixmap), QRect(QPoint(0, 0), size), mode, state)
        return pixmap



"""Module for QtWidgets."""
from __future__ import annotations

from collections.abc import Sequence

from ...qtpy.qt_compat import QT_API

if QT_API == "PySide6":
    from PySide6.QtWidgets import *  # type: ignore  # noqa: F403
elif QT_API == "PyQt6":
    from PyQt6.QtWidgets import *  # type: ignore  # noqa: F403
elif QT_API == "PyQt5":
    from PyQt5.QtWidgets import *  # type: ignore # noqa: F403
elif QT_API == "PySide2":
    from PySide2.QtWidgets import *  # type: ignore  # noqa: F403


class Application(QApplication):  # type: ignore  # noqa: F405
    """Override QApplication."""

    def __init__(self, args: Sequence[str] | None = None) -> None:
        """Override QApplication method."""
        super().__init__(args)

    def exec(self) -> int:
        """Override QApplication method."""
        if hasattr(super(), "exec"):
            return super().exec()
        return super().exec_()

    def exit(self, returnCode: int = 0) -> None:  # noqa: N803
        """Override QApplication method."""
        return super().exit(returnCode)


QApplication = Application



from astropy.io.fits.header import Header
import astropy.units as u
from astropy.units import Quantity, UnitConversionError
import numpy as np
from qtpy import QtGui, QtCore, QtWidgets
import pyqtgraph as pg

import abc
from dataclasses import asdict, dataclass, field
from enum import Enum, auto
import logging

from ..config import config, data_widgets
from ..config import SpectralLineData
from ..io.catalog import Catalog
from ..io.inspection_data import InspectionData, REDSHIFT_FILL_VALUE
from ..io.viewer_data import DataPath
from ..utils.widgets import AbstractWidget, MyViewBox

from .SmartSlider import SmartSlider

logger = logging.getLogger(__name__)


@dataclass
class Axis:
    unit: u.Unit | None = None
    scale: str = 'linear'
    label: str | None = None
    limits: tuple[float, float] = (0, 1)
    padding: float = 0

    @property
    def label_ext(self) -> str | None:
        if not self.label:
            return None

        unit_str = None
        if self.unit:
            unit_str = self.unit.to_string('unicode', fraction=True)

        label = self.label
        if self.scale == 'log':
            if unit_str:
                label = f'{label}/{unit_str}'
            label = f'log {label}'
        elif unit_str:
            label = f'{label} [{unit_str}]'

        return label

    @property
    def limits_padded(self) -> tuple[float, float]:
        w = self.limits[1] - self.limits[0]
        pad_abs = self.padding * w
        return self.limits[0] - pad_abs, self.limits[1] + pad_abs


@dataclass
class Axes:
    x: Axis = field(default_factory=Axis)
    y: Axis = field(default_factory=Axis)


class PlotTransformBase:
    def __init__(self, widget_title: str):
        self.widget_title = widget_title

    @abc.abstractmethod
    def apply(self, plot_data: Quantity | np.ndarray) -> Quantity | np.ndarray:
        pass


class ScaleTransform(PlotTransformBase):
    ALLOWED_PLOT_SCALES: tuple[str] = ('linear', 'log')

    def __init__(self, widget_title: str, scale='linear'):
        super().__init__(widget_title=widget_title)

        if scale not in self.ALLOWED_PLOT_SCALES:
            logger.error(f'Unknown scaling type: `{scale}` (widget: {self.widget_title})')
            self.scale = None

        self.scale = scale

    def apply(self, plot_data: Quantity | np.ndarray) -> Quantity | np.ndarray:
        if self.scale == 'log':
            if isinstance(plot_data, Quantity):
                plot_data = plot_data.value
            with np.errstate(invalid='ignore', divide='ignore'):
                plot_data = np.log10(plot_data)

        return plot_data


class UnitTransform(PlotTransformBase):
    def __init__(self, widget_title: str, unit: u.Unit):
        super().__init__(widget_title=widget_title)
        self.unit = unit

    def apply(self, plot_data: Quantity | np.ndarray) -> Quantity:
        if isinstance(plot_data, Quantity):
            try:
                plot_data = plot_data.value * plot_data.unit.to(self.unit)
            except UnitConversionError as e:
                logger.error(f'{e}. Axis unit will be ignored (widget: {self.widget_title})')
        else:
            plot_data = plot_data * self.unit

        return plot_data


# TODO: create Enum dynamically from SliderItem
class LinkableItem(Enum):
    XAXIS = auto()
    YAXIS = auto()
    COLORBAR = auto()
    S_SLIDER = auto()
    S_REDSHIFT = auto()


class SliderItem(Enum):
    SMOOTHING = auto()
    REDSHIFT = auto()


class ViewerElement(AbstractWidget):
    object_loaded = QtCore.Signal(str)
    object_destroyed = QtCore.Signal(str)

    allowed_data_types: tuple[type] | None = None

    def __init__(self, title: str, cfg: data_widgets.ViewerElement, appearance: config.Appearance,
                 spectral_lines: SpectralLineData, parent=None):
        self.title = title
        self.cfg = cfg
        self.appearance = appearance

        self.data_path: DataPath | None = None
        self.data = None
        self.meta: dict | Header | None = None

        self._object_loaded: bool = False

        self._graphics_view: pg.GraphicsView | None = None
        self.graphics_layout: pg.GraphicsLayout | None = None

        # graphics view properties
        self._axes: Axes | None = None
        self._qtransform: QtGui.QTransform | None = None

        # graphics items
        self.container: pg.PlotItem | None = None
        self._registered_items: list[pg.GraphicsItem] = []

        self._spectral_lines = spectral_lines
        self._spectral_line_artists: dict[str, tuple[pg.InfiniteLine, pg.TextItem]] = {}

        self.sliders: dict[SliderItem, SmartSlider] = {}
        self.smoothing_slider: SmartSlider | None = None
        self.redshift_slider: SmartSlider | None = None

        super().__init__(parent=parent)

        self.init_view()
        self.setEnabled(False)

    def _apply_axis_data_transform(self, plot_data: Quantity | np.ndarray, scale: str,
                                   unit: u.Unit | None) -> Quantity | np.ndarray:
        # order is important: first convert units, then apply scaling
        if unit:
            plot_data = UnitTransform(self.title, unit=unit).apply(plot_data)
        plot_data = ScaleTransform(self.title, scale=scale).apply(plot_data)

        return plot_data

    def apply_xdata_transform(self, plot_data: Quantity | np.ndarray) -> Quantity | np.ndarray:
        return self._apply_axis_data_transform(plot_data, scale=self._axes.x.scale, unit=self._axes.x.unit)

    def apply_ydata_transform(self, plot_data: Quantity | np.ndarray) -> Quantity | np.ndarray:
        return self._apply_axis_data_transform(plot_data, scale=self._axes.y.scale, unit=self._axes.y.unit)

    def _create_line_artists(self):
        if self._spectral_line_artists:
            for line_artist in self._spectral_line_artists.values():
                line_artist[0].deleteLater()
                line_artist[1].deleteLater()

        self._spectral_line_artists = {}

        line_color = self.cfg.spectral_lines.color
        if line_color is None:
            if self.appearance.theme == 'dark':
                line_color = (255, 255, 255)
            else:
                line_color = (175.68072, 220.68924, 46.59488)
        line_pen = pg.mkPen(color=line_color, width=1)

        for line_name, lambda0 in self._spectral_lines.wavelengths.items():
            line = pg.InfiniteLine(pen=line_pen)
            line.setZValue(10)

            label = pg.TextItem(text=line_name, color=line_color, anchor=(0, 1), angle=-90)
            label.setZValue(11)

            self._spectral_line_artists[line_name] = (line, label)

    def _add_line_artists(self):
        if self.cfg.spectral_lines.visible:
            for line_name, line_artist in self._spectral_line_artists.items():
                self.container.addItem(line_artist[0], ignoreBounds=True)
                self.container.addItem(line_artist[1])

    def init_ui(self):
        self._graphics_view = pg.GraphicsView(parent=self)

        self.graphics_layout = pg.GraphicsLayout()
        self._graphics_view.setCentralItem(self.graphics_layout)

        self.container = pg.PlotItem(name=self.title, viewBox=MyViewBox(enableMenu=False))
        self.set_axes_visibility()
        self.container.hideButtons()
        self.container.setMouseEnabled(True, True)

        self._create_line_artists()
        self._add_line_artists()

        self.graphics_layout.addItem(self.container, 0, 0)

        self._create_sliders()

    def _create_sliders(self):
        self.smoothing_slider = SmartSlider(short_name='sigma', action='smooth the data', parent=self,
                                            **asdict(self.cfg.smoothing_slider))
        self.redshift_slider = SmartSlider(short_name='z', full_name='redshift', parent=self,
                                           **asdict(self.cfg.redshift_slider))

        self.sliders[SliderItem.SMOOTHING] = self.smoothing_slider
        self.sliders[SliderItem.REDSHIFT] = self.redshift_slider

        self.smoothing_slider.value_changed[float].connect(self.smooth)
        self.redshift_slider.value_changed[float].connect(self.redshift_changed_action)

    def set_geometry(self, spacing: int, margins: int | tuple[int, int, int, int]):
        super().set_geometry(spacing=spacing, margins=margins)

        self.graphics_layout.setSpacing(spacing)
        self.graphics_layout.setContentsMargins(0, 0, 5, 5)

    def set_layout(self):
        self.setLayout(QtWidgets.QGridLayout())
        self.set_geometry(spacing=self.appearance.viewer_spacing, margins=self.appearance.viewer_margins)

    def populate(self):
        # add vertical sliders and the graphics view
        sub_layout = QtWidgets.QHBoxLayout()
        for s in self.sliders.values():
            if not s.show_text_editor:
                sub_layout.addWidget(s)
        sub_layout.addWidget(self._graphics_view)
        self.layout().addLayout(sub_layout, 1, 1, 1, 1)

        # add horizontal sliders
        sub_layout = QtWidgets.QVBoxLayout()
        for s in self.sliders.values():
            if s.show_text_editor:
                sub_layout.addWidget(s)
        self.layout().addLayout(sub_layout, 2, 1, 1, 1)

    def init_view(self):
        x_unit = u.Unit(self.cfg.x_axis.unit) if self.cfg.x_axis.unit else None
        y_unit = u.Unit(self.cfg.y_axis.unit) if self.cfg.y_axis.unit else None
        self._axes = Axes(x=Axis(unit=x_unit, scale=self.cfg.x_axis.scale, label=self.cfg.x_axis.label),
                          y=Axis(unit=y_unit, scale=self.cfg.y_axis.scale, label=self.cfg.y_axis.label))
        self._qtransform = QtGui.QTransform()

    def set_axes_visibility(self):
        self.container.showAxes((self.cfg.y_axis.visible, False, False, self.cfg.x_axis.visible),
                                showValues=(self.cfg.y_axis.visible, False, False, self.cfg.x_axis.visible))

    def update_axis_labels(self):
        show_xaxis = self.container.getAxis('bottom').isVisible()
        show_yaxis = self.container.getAxis('left').isVisible()

        self.container.setLabel(axis='bottom', text=self._axes.x.label_ext)
        self.container.setLabel(axis='left', text=self._axes.y.label_ext)

        # pyqtgraph changes axis visibility after adding labels
        self.container.showAxes((show_yaxis, False, False, show_xaxis))

    def setEnabled(self, a0: bool = True):
        super().setEnabled(a0)
        for line_artist in self._spectral_line_artists.values():
            line_artist[0].setVisible(a0)
            line_artist[1].setVisible(a0)
        for s in self.sliders.values():
            s.setEnabled(a0)

    @QtCore.Slot()
    def load_project(self):
        if self._object_loaded:
            self._destroy_object()

    @QtCore.Slot(object, object, object)
    def set_data(self, data, meta: dict | Header | None, data_path: DataPath | None):
        if data is not None and data_path is None:
            logger.error(f"Failed to set the widget data: data path not provided (widget: {self.title})")
            return

        self.data, self.meta, self.data_path = data, meta, data_path
        if self.data is None:
            self.meta, self.data_path = None, None

    @QtCore.Slot(int, InspectionData, object)
    def load_object(self, j: int, review: InspectionData, cat_entry: Catalog | None):
        if self._object_loaded:
            self._destroy_object()

        if self.data is None:
            return

        self.add_content()
        self.setup_view(cat_entry)
        self.setup_slider_view(j, review, cat_entry)
        self.setEnabled(True)

        self._object_loaded = True
        self.object_loaded.emit(self.title)

    def _destroy_object(self):
        self.clear_content()
        self.clear_view()
        self.clear_slider_view()
        self.setEnabled(False)

        self._object_loaded = False
        self.object_destroyed.emit(self.title)

    @abc.abstractmethod
    def add_content(self):
        pass

    def register_item(self, item: pg.GraphicsItem, **kwargs):
        item.setTransform(self._qtransform)
        self.container.addItem(item, **kwargs)
        self._registered_items.append(item)

    def setup_view(self, cat_entry: Catalog | None):
        xlim = (self.cfg.x_axis.limits.min, self.cfg.x_axis.limits.max)
        ylim = (self.cfg.y_axis.limits.min, self.cfg.y_axis.limits.max)

        xlim = (xlim[0] if xlim[0] is not None else self._axes.x.limits[0],
                xlim[1] if xlim[1] is not None else self._axes.x.limits[1])

        ylim = (ylim[0] if ylim[0] is not None else self._axes.y.limits[0],
                ylim[1] if ylim[1] is not None else self._axes.y.limits[1])

        self.set_default_range(xrange=xlim, yrange=ylim)

        self.update_axis_labels()
        self.apply_qtransform()

    def setup_slider_view(self, j: int, review: InspectionData, cat_entry: Catalog | None):
        for slider_name, s in self.sliders.items():
            s.set_default_value_from_catalog(cat_entry)  # load the catalog value to the slider
            if slider_name is SliderItem.REDSHIFT:  # load redshift from inspection results
                redshift = review.get_value(j, 'z_sviz')
                if not np.isclose(redshift, REDSHIFT_FILL_VALUE):
                    s.set_default_value(redshift)

    def set_default_range(self, xrange: tuple[float, float] | None = None, yrange: tuple[float, float] | None = None,
                          apply_qtransform=False):
        if apply_qtransform:
            if not xrange or not yrange:
                raise ValueError("Can only apply transformation when both x- and y-axis limits are specified")

            x1, y1 = self._qtransform.map(xrange[0], yrange[0])
            x2, y2 = self._qtransform.map(xrange[1], yrange[1])

            xrange, yrange = (x1, x2), (y1, y2)

        if xrange:
            self._axes.x.limits = xrange
        if yrange:
            self._axes.y.limits = yrange

    def set_content_padding(self, xpad: float | None = None, ypad: float | None = None):
        if xpad:
            self._axes.x.padding = xpad
        if ypad:
            self._axes.y.padding = ypad

    def apply_qtransform(self, apply_to_default_range=False):
        if apply_to_default_range:
            self.set_default_range(self._axes.x.limits, self._axes.y.limits, apply_qtransform=True)

        for item in self._registered_items:
            item.setTransform(self._qtransform)

    def smooth(self, sigma: float):
        pass

    def set_spectral_line_positions(self, redshift: float = 0):
        # check unit compatibility
        line_unit = u.Unit(self._spectral_lines.wave_unit)
        if self._axes.x.unit:
            try:
                line_unit.to(self._axes.x.unit)
            except UnitConversionError as e:
                if self.cfg.spectral_lines.visible:
                    logger.error(f'Failed to calculate positions of spectral lines: {e} (widget: {self.title})')
                return

        scale0 = 1 + redshift
        y_min, y_max = self._axes.y.limits
        label_height = y_max - (y_max - y_min) * 0.03

        line_waves = np.array([self._spectral_lines.wavelengths[line_name]
                               for line_name in self._spectral_line_artists.keys()]) * scale0
        line_waves = self.apply_xdata_transform(line_waves * u.Unit('AA'))
        if isinstance(line_waves, Quantity):
            line_waves = line_waves.value

        for line_wave, line_artist in zip(line_waves, self._spectral_line_artists.values()):
            line_artist[0].setPos(line_wave)
            line_artist[1].setPos(QtCore.QPointF(line_wave, label_height))

    def redshift_changed_action(self, redshift: float):
        self.set_spectral_line_positions(redshift)

    def reset_range(self, x_axis_links: dict | None = None, y_axis_links: dict | None = None):
        x_range, y_range = self._axes.x.limits_padded, self._axes.y.limits_padded
        if x_axis_links and self.title in x_axis_links:
            x_range = None
        if y_axis_links and self.title in y_axis_links:
            y_range = None
        self.container.setRange(xRange=x_range, yRange=y_range, padding=0)

    @QtCore.Slot(object)
    def reset_view(self, widget_links: dict | None = None):
        if self.data is None:
            return
        if widget_links is None:
            widget_links = dict()

        self.reset_range(widget_links.get(LinkableItem.XAXIS), widget_links.get(LinkableItem.YAXIS))
        if self.title not in widget_links.get(LinkableItem.S_REDSHIFT, dict()):
            self.redshift_slider.reset()
        if self.title not in widget_links.get(LinkableItem.S_SLIDER, dict()):
            self.smoothing_slider.update_from_slider()  # only update, do not reset

    def remove_registered_items(self):
        while self._registered_items:
            item = self._registered_items.pop()
            self.container.removeItem(item)

    def clear_content(self):
        self.remove_registered_items()

    def clear_view(self):
        self.init_view()

    def clear_slider_view(self):
        for s in self.sliders.values():
            s.clear()

    def get_dock_title(self) -> str:
        if self.data_path is None or self.cfg.dock_title_fmt == 'fixed':
            return self.title

        if self.meta is None:
            return self.data_path.name

        # adding EXTNAME and EXTVER to the dock title
        fits_meta = self.meta.get('EXTNAME'), self.meta.get('EXTVER')
        j = 0
        while j < len(fits_meta) and fits_meta[j] is not None:
            j += 1
        title_meta = ', '.join(map(str, fits_meta[:j]))

        if not title_meta:
            return self.data_path.name

        if self.cfg.dock_title_fmt == 'short':
            return title_meta  # str(fits_meta[j - 1])

        return f"{self.data_path.name} [{title_meta}]"

    def _enter_zen_mode(self):
        self.container.showAxes(False, showValues=False)
        self.smoothing_slider.setVisible(False)

    def _update_visibility(self):
        self.set_axes_visibility()
        self.smoothing_slider.setVisible(self.cfg.smoothing_slider.visible)

    @QtCore.Slot(bool)
    def update_visibility(self, is_zen: bool):
        self._update_visibility()

        if is_zen:
            self._enter_zen_mode()

    @QtCore.Slot()
    def update_spectral_lines(self):
        self._create_line_artists()
        self._add_line_artists()
